#include "GLWindow.h"
#include <iostream>
#include "ngl/Camera.h"
#include "ngl/Transformation.h"
#include "ngl/TransformStack.h"
#include "ngl/NGLInit.h"
#include "ngl/ShaderManager.h"
#include "ngl/Util.h"
#include "ngl/VertexArrayObject.h"
#include <QImage>
#include <ctime>

/// On the mac these are extentions, it's easier to let the
/// pre-processor deal with this than using GLEW / other code
/// as on Win and Linux it works as normal glGen




//----------------------------------------------------------------------------------------------------------------------
GLWindow::GLWindow(
                   QWidget *_parent
                  ) :
                    QGLWidget(_parent)
{

	// set this widget to have the initial keyboard focus
	setFocus();
	// re-size the widget to that of the parent (in this case the GLFrame passed in on construction)
	this->resize(_parent->size());
	// Now set the initial GLWindow attributes to default values
	// Roate is false
	m_rotate=false;
	// mouse rotation values set to 0
	m_spinXFace=0;
	m_spinYFace=0;
	m_origX=0;
	m_origY=0;
}

GLWindow::~GLWindow()
{
  ngl::NGLInit *Init = ngl::NGLInit::instance();
  std::cout<<"Shutting down NGL, removing VBO's and Shaders\n";
  Init->NGLQuit();
}


void GLWindow::buildVAO()
{
	// create a vao as a series of GL_QUADS
	m_vao= ngl::VertexArrayObject::createVOA(GL_QUADS);
	m_vao->bind();


   GLfloat vertices[] = {
                         1, 1, 1, -1, 1, 1, -1,-1, 1,  1,-1, 1,  // v0-v1-v2-v3
                         1, 1, 1,  1,-1, 1,  1,-1,-1,  1, 1,-1,  // v0-v3-v4-v5
                         1, 1, 1,  1, 1,-1, -1, 1,-1, -1, 1, 1,  // v0-v5-v6-v1
                        -1, 1, 1, -1, 1,-1, -1,-1,-1, -1,-1, 1,  // v1-v6-v7-v2
                        -1,-1,-1,  1,-1,-1,  1,-1, 1, -1,-1, 1,  // v7-v4-v3-v2
                         1,-1,-1, -1,-1,-1, -1, 1,-1,  1, 1,-1   // v4-v7-v6-v5
                        };
	 // in this case we are going to set our data as the vertices above
	m_vao->setData(24*3,vertices[0]);
	// now we set the attribute pointer to be 0 (as this matches vertIn in our shader)
	m_vao->setVertexAttributePointer(0,3,GL_FLOAT,0,0);
  GLfloat texture[] = {
												0,0,  0,1, 1,1, 1,0,  // v0-v1-v2-v3
												0,0,  0,1, 1,1, 1,0,  // v0-v1-v2-v3
												0,0,  0,1, 1,1, 1,0,  // v0-v1-v2-v3
												0,0,  0,1, 1,1, 1,0,  // v0-v1-v2-v3
												0,0,  0,1, 1,1, 1,0,  // v0-v1-v2-v3
												0,0,  0,1, 1,1, 1,0,  // v0-v1-v2-v3
											};
	// We can now create another set of data (which will be added to the VAO)
	// in this case the UV co-ords
	m_vao->setData(24*2,texture[0]);
	// now we set this as the 2nd attribute pointer (1) to match inUV in the shader
	m_vao->setVertexAttributePointer(1,2,GL_FLOAT,0,0);
	// tell the VAO how many verts to draw
	m_vao->setNumIndices(24);
	// now unbind
	m_vao->unbind();

}

// This virtual function is called once before the first call to paintGL() or resizeGL(),
//and then once whenever the widget has been assigned a new QGLContext.
// This function should set up any required OpenGL context rendering flags, defining display lists, etc.

//----------------------------------------------------------------------------------------------------------------------
void GLWindow::initializeGL()
{
	ngl::NGLInit *Init = ngl::NGLInit::instance();
	#ifdef WIN32
		glewInit();
	#endif
	Init->initGlew();

	glClearColor(0.4f, 0.4f, 0.4f, 1.0f);			   // Grey Background
	// enable depth testing for drawing
	glEnable(GL_DEPTH_TEST);
	glEnable(GL_MULTISAMPLE);
	// Now we will create a basic Camera from the graphics library
	// This is a static camera so it only needs to be set once
	// First create Values for the camera position
	ngl::Vector from(2,1,2);
	ngl::Vector to(0,0,0);
	ngl::Vector up(0,1,0);

	m_cam= new ngl::Camera(from,to,up,ngl::PERSPECTIVE);
	// set the shape using FOV 45 Aspect Ratio based on Width and Height
	// The final two are near and far clipping planes of 0.5 and 10
	m_cam->setShape(45,(float)720.0/576.0,0.001,150,ngl::PERSPECTIVE);
	// now to load the shader and set the values
	// grab an instance of shader manager
	ngl::ShaderManager *shader=ngl::ShaderManager::instance();
	// load a frag and vert shaders

	shader->createShaderProgram("TextureShader");

	shader->attachShader("SimpleVertex",ngl::VERTEX);
	shader->attachShader("SimpleFragment",ngl::FRAGMENT);
	shader->loadShaderSource("SimpleVertex","shaders/Vertex.vs");
	shader->loadShaderSource("SimpleFragment","shaders/Fragment.fs");

	shader->compileShader("SimpleVertex");
	shader->compileShader("SimpleFragment");
	shader->attachShaderToProgram("TextureShader","SimpleVertex");
	shader->attachShaderToProgram("TextureShader","SimpleFragment");

	shader->bindAttribute("TextureShader",0,"inVert");
	shader->bindAttribute("TextureShader",1,"inUV");

	shader->linkProgramObject("TextureShader");
	(*shader)["TextureShader"]->use();
	// now pass the modelView and projection values to the shader
	shader->setShaderParamFromMatrix("TextureShader","ViewMatrix",m_cam->getModelView());
	shader->setShaderParamFromMatrix("TextureShader","projectionMatrix",m_cam->getProjection());
	buildVAO();
	ngl::Texture t("textures/crate.bmp");
	t.setTextureGL();
}

//----------------------------------------------------------------------------------------------------------------------
//This virtual function is called whenever the widget has been resized.
// The new size is passed in width and height.
void GLWindow::resizeGL(
                        int _w,
                        int _h
                       )
{

  glViewport(0,0,_w,_h);
  m_cam->setShape(45,(float)_w/_h,0.001,350,ngl::PERSPECTIVE);
  ngl::ShaderManager *shader=ngl::ShaderManager::instance();
  (*shader)["TextureShader"]->use();
	shader->setShaderParamFromMatrix("TextureShader","projectionMatrix",m_cam->getProjection());
}

//----------------------------------------------------------------------------------------------------------------------
//This virtual function is called whenever the widget needs to be painted.
// this is our main drawing routine
void GLWindow::paintGL()
{
  // clear the screen and depth buffer
  glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT);
 // build our transform stack
  ngl::Transformation trans;
  // set the mouse rotation
  trans.setRotation(m_spinXFace,m_spinYFace,0);
  // set this in the TX stack
  m_transformStack.setGlobal(trans);

  ngl::ShaderManager *shader=ngl::ShaderManager::instance();
	(*shader)["TextureShader"]->use();
  m_transformStack.loadGlobalAndCurrentMatrixToShader("TextureShader","ModelMatrix");

	shader->setShaderParamFromMatrix("TextureShader","ModelMatrix",m_transformStack.getCurrAndGlobal().getMatrix());
	// now we bind back our vertex array object and draw
	m_vao->bind();
	int instances=0;
	for (float z=-8; z<8; z+=0.5)
	{
	  for (float x=-8; x<8; x+=0.5)
	  {
			m_transformStack.pushTransform();
			{
				m_transformStack.getCurrentTransform().setRotation(x*20,(x*z)*40,z*2);
				m_transformStack.getCurrentTransform().setPosition(ngl::Vector(x,0.49,z));
				m_transformStack.setScale(0.2,0.2,0.2);
				shader->setShaderParamFromMatrix("TextureShader","ModelMatrix",m_transformStack.getCurrAndGlobal().getMatrix());
				++instances;
				m_vao->draw();
			} // and before a pop
			 m_transformStack.popTransform();
	  }
	}
	m_vao->unbind();

}

//----------------------------------------------------------------------------------------------------------------------
void GLWindow::mouseMoveEvent (
                               QMouseEvent * _event
                              )
{
  // note the method buttons() is the button state when event was called
  // this is different from button() which is used to check which button was
  // pressed when the mousePress/Release event is generated
  if(m_rotate && _event->buttons() == Qt::LeftButton)
  {
    m_spinYFace = ( m_spinYFace + (_event->x() - m_origX) ) % 360 ;
    m_spinXFace = ( m_spinXFace + (_event->y() - m_origY) ) % 360 ;
    m_origX = _event->x();
    m_origY = _event->y();
  }
  // re-draw GL
  updateGL();
}


//----------------------------------------------------------------------------------------------------------------------
void GLWindow::mousePressEvent (
                                QMouseEvent * _event
                               )
{
  // this method is called when the mouse button is pressed in this case we
  // store the value where the maouse was clicked (x,y) and set the Rotate flag to true
  if(_event->button() == Qt::LeftButton)
  {
    m_origX = _event->x();
    m_origY = _event->y();
    m_rotate =true;
  }
}

//----------------------------------------------------------------------------------------------------------------------
void GLWindow::mouseReleaseEvent (
                                  QMouseEvent * _event
                                 )
{
  // this event is called when the mouse button is released
  // we then set Rotate to false
  if (_event->button() == Qt::LeftButton)
  {
    m_rotate=false;
  }
}

