#include "GLWindow.h"
#include <iostream>
#include "ngl/Camera.h"
#include "ngl/Transformation.h"
#include "ngl/TransformStack.h"
#include "ngl/NGLInit.h"
#include "ngl/ShaderManager.h"
#include "ngl/Util.h"
#include "ngl/VertexArrayObject.h"
#include <ctime>



//----------------------------------------------------------------------------------------------------------------------
GLWindow::GLWindow(
                   QWidget *_parent
                  ) :
                    QGLWidget(_parent)
{

	// set this widget to have the initial keyboard focus
	setFocus();
	// re-size the widget to that of the parent (in this case the GLFrame passed in on construction)
	this->resize(_parent->size());
	// Now set the initial GLWindow attributes to default values
	// Roate is false
	m_rotate=false;
	// mouse rotation values set to 0
	m_spinXFace=0;
	m_spinYFace=0;
	m_origX=0;
	m_origY=0;

}

GLWindow::~GLWindow()
{
  ngl::NGLInit *Init = ngl::NGLInit::instance();
  std::cout<<"Shutting down NGL, removing VBO's and Shaders\n";
  Init->NGLQuit();
	// now we have finished remove the VOA
	m_vaoSphere->removeVOA();
}

// This virtual function is called once before the first call to paintGL() or resizeGL(),
//and then once whenever the widget has been assigned a new QGLContext.
// This function should set up any required OpenGL context rendering flags, defining display lists, etc.

//----------------------------------------------------------------------------------------------------------------------
void GLWindow::initializeGL()
{
	ngl::NGLInit *Init = ngl::NGLInit::instance();
	#ifdef WIN32
		glewInit();
	#endif
	Init->initGlew();

	glClearColor(0.4f, 0.4f, 0.4f, 1.0f);			   // Grey Background
	// enable depth testing for drawing
	glEnable(GL_DEPTH_TEST);
	glEnable(GL_MULTISAMPLE);
	// Now we will create a basic Camera from the graphics library
	// This is a static camera so it only needs to be set once
	// First create Values for the camera position
	ngl::Vector from(2,1,2);
	ngl::Vector to(0,0,0);
	ngl::Vector up(0,1,0);

	m_cam= new ngl::Camera(from,to,up,ngl::PERSPECTIVE);
	// set the shape using FOV 45 Aspect Ratio based on Width and Height
	// The final two are near and far clipping planes of 0.5 and 10
	m_cam->setShape(45,(float)720.0/576.0,0.001,150,ngl::PERSPECTIVE);
	// now to load the shader and set the values
	// grab an instance of shader manager
	ngl::ShaderManager *shader=ngl::ShaderManager::instance();
	// load a frag and vert shaders

	shader->createShaderProgram("TextureShader");

	shader->attachShader("SimpleVertex",ngl::VERTEX);
	shader->attachShader("SimpleFragment",ngl::FRAGMENT);
	shader->loadShaderSource("SimpleVertex","shaders/Vertex.vs");
	shader->loadShaderSource("SimpleFragment","shaders/Fragment.fs");

	shader->compileShader("SimpleVertex");
	shader->compileShader("SimpleFragment");
	shader->attachShaderToProgram("TextureShader","SimpleVertex");
	shader->attachShaderToProgram("TextureShader","SimpleFragment");

	shader->bindAttribute("TextureShader",0,"inVert");
	shader->bindAttribute("TextureShader",1,"inUV");
	shader->bindAttribute("TextureShader",2,"inNormal");


	shader->linkProgramObject("TextureShader");
	(*shader)["TextureShader"]->use();
	// now pass the modelView and projection values to the shader
	shader->setShaderParamFromMatrix("TextureShader","ViewMatrix",m_cam->getModelView());
	shader->setShaderParamFromMatrix("TextureShader","projectionMatrix",m_cam->getProjection());
	// build our VertexArrayObject
	buildVAOSphere();
	// load and set a texture
	ngl::Texture t("textures/earth.tiff");
	t.setTextureGL();
	glGenerateMipmap(GL_TEXTURE_2D);
	glEnable(GL_MULTISAMPLE);
}

//----------------------------------------------------------------------------------------------------------------------
//This virtual function is called whenever the widget has been resized.
// The new size is passed in width and height.
void GLWindow::resizeGL(
                        int _w,
                        int _h
                       )
{

  glViewport(0,0,_w,_h);
  m_cam->setShape(45,(float)_w/_h,0.001,350,ngl::PERSPECTIVE);
  ngl::ShaderManager *shader=ngl::ShaderManager::instance();
  (*shader)["TextureShader"]->use();
	shader->setShaderParamFromMatrix("TextureShader","projectionMatrix",m_cam->getProjection());
}

// a simple structure to hold our vertex data
struct vertData
{
  GLfloat u;
	GLfloat v;
	GLfloat nx;
	GLfloat ny;
	GLfloat nz;
	GLfloat x;
	GLfloat y;
	GLfloat z;
};

void GLWindow::buildVAOSphere()
{
	//  Sphere code based on a function Written by Paul Bourke.
	//  http://astronomy.swin.edu.au/~pbourke/opengl/sphere/
	// first we grab an instance of our VOA class as a TRIANGLE_STRIP
	m_vaoSphere= ngl::VertexArrayObject::createVOA(GL_TRIANGLE_STRIP);
	// next we bind it so it's active for setting data
	m_vaoSphere->bind();
	// the next part of the code calculates the P,N,UV of the sphere for tri_strips
	int buffSize;

	float theta1 = 0.0;
	float theta2 = 0.0;
	float theta3 = 0.0;


	float radius=1.0;
	float precision=80;
	// a std::vector to store our verts, remember vector packs contiguously so we can use it
	std::vector <vertData> data;
	// calculate how big our buffer is
	buffSize = (precision/2) * ((precision+1)*2);
	// Disallow a negative number for radius.
	if( radius < 0 )
	{
	radius = -radius;
	}
	// Disallow a negative number for _precision.
	if( precision < 4 )
	{
	precision = 4;
	}
	// now fill in a vertData structure and add to the data list for our sphere
	vertData d;
	for( int i = 0; i < precision/2; ++i )
	{
		theta1 = i * ngl::TWO_PI / precision - ngl::PI2;
		theta2 = (i + 1) * ngl::TWO_PI / precision - ngl::PI2;

		for( int j = 0; j <= precision; ++j )
		{
			theta3 = j * ngl::TWO_PI / precision;

			d.nx = cosf(theta2) * cosf(theta3);
			d.ny = sinf(theta2);
			d.nz = cosf(theta2) * sinf(theta3);
			d.x = radius * d.nx;
			d.y = radius * d.ny;
			d.z = radius * d.nz;

			d.u  = (j/(float)precision);
			d.v  = 2*(i+1)/(float)precision;

			data.push_back(d);

			d.nx = cosf(theta1) * cosf(theta3);
			d.ny = sinf(theta1);
			d.nz = cosf(theta1) * sinf(theta3);
			d.x = radius * d.nx;
			d.y = radius * d.ny;
			d.z = radius * d.nz;

			d.u  = (j/(float)precision);
			d.v  = 2*i/(float)precision;
			data.push_back(d);
		} // end inner loop
	}// end outer loop


	// now we have our data add it to the VAO, we need to tell the VAO the following
	// how much (in bytes) data we are copying
	// a pointer to the first element of data (in this case the address of the first element of the
	// std::vector
	m_vaoSphere->setData(buffSize*sizeof(vertData),data[0].u);
	// in this case we have packed our data in interleaved format as follows
	// u,v,nx,ny,nz,x,y,z
	// If you look at the shader we have the following attributes being used
	// attribute vec3 inVert; attribute 0
	// attribute vec2 inUV; attribute 1
	// attribute vec3 inNormal; attribure 2
	// so we need to set the vertexAttributePointer so the correct size and type as follows
	// vertex is attribute 0 with x,y,z(3) parts of type GL_FLOAT, our complete packed data is
	// sizeof(vertData) and the offset into the data structure for the first x component is 5 (u,v,nx,ny,nz)..x
	m_vaoSphere->setVertexAttributePointer(0,3,GL_FLOAT,sizeof(vertData),5);
	// uv same as above but starts at 0 and is attrib 1 and only u,v so 2
	m_vaoSphere->setVertexAttributePointer(1,2,GL_FLOAT,sizeof(vertData),0);
	// normal same as vertex only starts at position 2 (u,v)-> nx
	m_vaoSphere->setVertexAttributePointer(2,3,GL_FLOAT,sizeof(vertData),2);
	// now we have set the vertex attributes we tell the VAO class how many indices to draw when
	// glDrawArrays is called, in this case we use buffSize (but if we wished less of the sphere to be drawn we could
	// specify less (in steps of 3))
	m_vaoSphere->setNumIndices(buffSize);
	// finally we have finished for now so time to unbind the VAO
	m_vaoSphere->unbind();
}


//----------------------------------------------------------------------------------------------------------------------
//This virtual function is called whenever the widget needs to be painted.
// this is our main drawing routine
void GLWindow::paintGL()
{
  // clear the screen and depth buffer
  glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT);
 // build our transform stack
  ngl::Transformation trans;
  // set the mouse rotation
  trans.setRotation(m_spinXFace,m_spinYFace,0);
  // set this in the TX stack
  m_transformStack.setGlobal(trans);

  ngl::ShaderManager *shader=ngl::ShaderManager::instance();
	(*shader)["TextureShader"]->use();
  m_transformStack.loadGlobalAndCurrentMatrixToShader("TextureShader","ModelMatrix");

	shader->setShaderParamFromMatrix("TextureShader","ModelMatrix",m_transformStack.getCurrAndGlobal().getMatrix());
	// now we bind back our vertex array object and draw
	m_vaoSphere->bind();
	m_vaoSphere->draw();

	// now we are done so unbind
	m_vaoSphere->unbind();
}

//----------------------------------------------------------------------------------------------------------------------
void GLWindow::mouseMoveEvent (
                               QMouseEvent * _event
                              )
{
  // note the method buttons() is the button state when event was called
  // this is different from button() which is used to check which button was
  // pressed when the mousePress/Release event is generated
  if(m_rotate && _event->buttons() == Qt::LeftButton)
  {
    m_spinYFace = ( m_spinYFace + (_event->x() - m_origX) ) % 360 ;
    m_spinXFace = ( m_spinXFace + (_event->y() - m_origY) ) % 360 ;
    m_origX = _event->x();
    m_origY = _event->y();
  }
  // re-draw GL
  updateGL();
}


//----------------------------------------------------------------------------------------------------------------------
void GLWindow::mousePressEvent (
                                QMouseEvent * _event
                               )
{
  // this method is called when the mouse button is pressed in this case we
  // store the value where the maouse was clicked (x,y) and set the Rotate flag to true
  if(_event->button() == Qt::LeftButton)
  {
    m_origX = _event->x();
    m_origY = _event->y();
    m_rotate =true;
  }
}

//----------------------------------------------------------------------------------------------------------------------
void GLWindow::mouseReleaseEvent (
                                  QMouseEvent * _event
                                 )
{
  // this event is called when the mouse button is released
  // we then set Rotate to false
  if (_event->button() == Qt::LeftButton)
  {
    m_rotate=false;
  }
}

